/* drivers/barcelona/gpio/gpio.c
 *
 * Implementation of the TomTom Dock (GPIO) driver.
 *
 * Copyright (C) 2004,2005,2006 TomTom BV <http://www.tomtom.com/>
 * Authors: Dimitry Andric <dimitry.andric@tomtom.com>
 *          Koen Martens <kmartens@sonologic.nl>
 *          Jeroen Taverne <jeroen.taverne@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

/* Includes */
#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/wait.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/poll.h>
#include <linux/device.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/stat.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/irq.h>
#include <asm/hardware.h>
#include <asm/hardware/clock.h>
#include <asm/arch/regs-gpio.h>
#include <barcelona/Barc_Gpio.h>
#include <barcelona/gopins.h>
#include <barcelona/debug.h>
#include <barcelona/Barc_adc.h>
#include <barcelona/Barc_buspower.h>
#include "ns73.h"
#include "buspower.h"

/* Defines */
#define PFX "gpio: "
#define PK_DBG PK_DBG_FUNC

#define GPIO_POLL_DELAY (HZ / 5)

/* Forward declarations */
static void gpio_status_poll(unsigned long data);

/* Globals */
static HARDWARE_STATUS gpio_hw_status;
static unsigned int diskAccessStarted = 0;
static unsigned int btModeState;
static struct timer_list gpio_timer = TIMER_INITIALIZER(gpio_status_poll, 0, 0);
static int power_button_timer = -1;
static int ignition_timer = -1;
static int diskAccess = 0;
static atomic_t uart_break_cnt;
static int uart_low_cnt = 0;
static int uart_high_cnt = 0;
extern int bt_failed;
static wait_queue_head_t gpio_wait;
static int statusChanged = 0;
static unsigned int frequency = 92000000;
static unsigned int oldFrequency = 0;
static int cycleDockPowerTimer = 0;

static void gpio_tasklet(unsigned long data);
DECLARE_TASKLET(gpio_tasklet_handle, gpio_tasklet, 0L);

/* Functions */
void gpio_handle_break(int uart)
{
		/* only support for UART0 */
		if (uart == 0) 
		{
			atomic_inc(&uart_break_cnt);
		}
		return;
}

static void gpio_tasklet(unsigned long data)
{
	static unsigned int oldInputStatus = 0;

	// Only do dock power cycle if dock power is available
	if (IO_HaveDockPower())
	{
		// Check if power cycle is in progress
		if (cycleDockPowerTimer > 0)
		{
			// Deactivate TMC power and UART signals so all signals on TMC connector are low
			if (!(gpio_hw_status.u8InputStatus & DOCK_FM_TRANSMITTER_MASK)) IO_Deactivate(DOCK_PWREN);
			IO_Deactivate(TMC_POWER);
			IO_Deactivate(RXD_DOCK);
			IO_Deactivate(TXD_DOCK);
		}
		else
		if (cycleDockPowerTimer == 0)
		{
			cycleDockPowerTimer = -1;
			// Enable TMC power and UART signals again and wait a while
			IO_Activate(DOCK_PWREN);
			IO_Activate(TMC_POWER);
			IO_SetFunction(RXD_DOCK);
			IO_SetFunction(TXD_DOCK);
			mdelay(10);
		}
	}

	// Only access I2C on dock when possible
	if ((IO_GetInput(DOCK_PWREN)) && (IO_HaveIoExpander()))
	{
		// Update IO expander
		IO_Update();

		// Update Liverpool if detected
		if ((IO_GetDockState() == DOCK_WINDSCREEN) && (NS73_Detected()))
		{
			gpio_hw_status.u8InputStatus |= DOCK_FM_TRANSMITTER_MASK;
			if (frequency != oldFrequency)
			{
				NS73_SetFrequency(frequency);
				oldFrequency = frequency;
			}
		}
		else
		{
			gpio_hw_status.u8InputStatus &= ~DOCK_FM_TRANSMITTER_MASK;
			// Force update of transmittnext time it's detected
			oldFrequency = 0;
		}
	}
	else
	{
		gpio_hw_status.u8InputStatus &= ~DOCK_FM_TRANSMITTER_MASK;
		// Force update of transmittnext time it's detected
		oldFrequency = 0;
	}

	gpio_hw_status.u8DockStatus = IO_GetDockState();

	if (IO_GetInput(LIGHTS_DETECT)) gpio_hw_status.u8InputStatus |= DOCK_LIGHTS_MASK; else gpio_hw_status.u8InputStatus &= ~DOCK_LIGHTS_MASK;
	if (IO_GetInput(IGNITION)) gpio_hw_status.u8InputStatus |= DOCK_IGNITION_MASK; else gpio_hw_status.u8InputStatus &= ~DOCK_IGNITION_MASK;
	if (IO_GetInput(EXTMIC_DETECT)) gpio_hw_status.u8InputStatus |= DOCK_EXTMIC_MASK; else gpio_hw_status.u8InputStatus &= ~DOCK_EXTMIC_MASK;
	if (IO_GetInput(HEADPHONE_DETECT)) gpio_hw_status.u8InputStatus |= DOCK_HEADPHONE_MASK; else gpio_hw_status.u8InputStatus &= ~DOCK_HEADPHONE_MASK;
	if (IO_GetInput(LINEIN_DETECT)) gpio_hw_status.u8InputStatus |= DOCK_LINEIN_MASK; else gpio_hw_status.u8InputStatus &= ~DOCK_LINEIN_MASK;
	if (!IO_HaveUsbBusPowered()) 
	{
		/* simple host detection, just VBUS, needed for valencia and go700 since have no S3C USB reset detect circuitry */
		if (IO_GetInput(USB_HOST_DETECT)) gpio_hw_status.u8InputStatus |= USB_DETECT_MASK; else gpio_hw_status.u8InputStatus &= ~USB_DETECT_MASK;
	} 
	else 
	{
		/* host detection with USB reset signal */
		buspower_poll();
#ifdef SUPPORT_FAKE_DEVICE
		/* wait on configured device */
		if (buspower_ModeDetect() == USB_MODE_DEVICE) gpio_hw_status.u8InputStatus |= USB_DETECT_MASK; else gpio_hw_status.u8InputStatus &= ~USB_DETECT_MASK;
#else
		/* wait on simple USB reset -> host detected */
		if (buspower_ModeDetect() == USB_MODE_HOST) gpio_hw_status.u8InputStatus |= USB_DETECT_MASK; else gpio_hw_status.u8InputStatus &= ~USB_DETECT_MASK;
#endif
	}
	
	/* UART RXD line test for valencia and other extended dock users - NEEDS PULLDOWN (2442) to work */
	/* optionally, for interupt-driven detection, a parallel interrupt input RXD_DOCK_INT to be there */
	if (IO_HaveRxdDetect()) 
	{
		if (gpio_hw_status.u8InputStatus & DOCK_RXD_DETECT_MASK) 
		{
			int break_cnt = 0;
			/* rxd was previously detected (high) */
			/* check if any breaks occurred in UART0 driver */
			while(atomic_sub_and_test(1, &uart_break_cnt)) 
			{
				break_cnt++;
			}
			if (break_cnt) 
			{
				IO_SetInput(RXD_DOCK);
				if (IO_GetInput(RXD_DOCK) == 0) 
				{
					/* uart still low, good */
					gpio_hw_status.u8InputStatus &= ~DOCK_RXD_DETECT_MASK;
				} 
				else 
				{
					/* break detected, but line still high - let it debounce, uart will count another break */
					IO_SetFunction(RXD_DOCK);
				}
			} 
			else 
			{
				/* no break counted, line assumed to be busy, most of the time high */
				if (IO_GetInput(RXD_DOCK) == 0) 
				{
					uart_low_cnt++;
					if (uart_low_cnt > 5) 
					{
						// ca 1 sec low, disconnected
						gpio_hw_status.u8InputStatus &= ~DOCK_RXD_DETECT_MASK;
					}
				} 
				else 
				{
					uart_low_cnt = 0;
				}
			}
		} 
		else 
		{
			/* rxd was low (disconnected), check if came up */
			IO_SetInput(RXD_DOCK);
			if (IO_GetInput(RXD_DOCK) == 0) 
			{
				/* still low, good */
				uart_high_cnt = 0;
			} 
			else 
			{
				/* line assumed to be busy or still high */
				if (uart_high_cnt < 5) 
				{
					uart_high_cnt++;
				} 
				else 
				{
					IO_SetFunction(RXD_DOCK);
					gpio_hw_status.u8InputStatus |= DOCK_RXD_DETECT_MASK;
				}
			}
		}
	} 
	else 
	{
		/* no serial detect dock */
		/* serial always active */
		gpio_hw_status.u8InputStatus |= DOCK_RXD_DETECT_MASK;
	}
	
	// Auto-switch off if in dock and power has been off for more than five seconds
	// Check if docked in car
	if (IO_CarDocked())
	{
		// External power?
		if (IO_GetInput(ACPWR)) 
		{
			// Enable running timer when external power is removed (wait 5 seconds)
			ignition_timer = 25;
		}
		else
		{
			// Count down timer when external power was active while docked
			if (ignition_timer > 0)
			{
				ignition_timer--;
			}
			else
			if (ignition_timer == 0)
			{
				// Disable timer to wait for connection of external power
				ignition_timer = -1;
				// Send message to application to shut down
				if (IO_HaveGpioAutoShutdown())
					gpio_hw_status.u8InputStatus |= ONOFF_MASK;
				else
					gpio_hw_status.u8InputStatus |= DOCK_ONOFF_MASK;
			}
		}
	}
	else 
	{
		// Not docked, so disable timer
		ignition_timer = -1;
	}
	
	// Button pressed?
	if (IO_GetInput(ON_OFF)) 
	{
		// Count down timer when button was released
		if (power_button_timer > 0)
		{
			power_button_timer--;
		}
		else
		if (power_button_timer == 0)
		{
			// Disable timer to wait for button release
			power_button_timer = -1;
			// Send message to application to shut down
			gpio_hw_status.u8InputStatus |= ONOFF_MASK;
		}
	}
	else
	{
		// Enable running timer when button is released (wait 1/5 second)
		power_button_timer = 1;
	}
	
	// HDD access?
	if (IO_GetInput(HDD_LED))
	{
		// Transition to disk access started?
		if (diskAccess == 0)
		{
			diskAccess = 1;
			diskAccessStarted = 1;
		}
	} else diskAccess = 0;
	
	if (oldInputStatus != gpio_hw_status.u8InputStatus)
	{
		oldInputStatus = gpio_hw_status.u8InputStatus;
		statusChanged = 1;
		wake_up_interruptible(&gpio_wait);
	}
}

static void gpio_status_poll(unsigned long data)
{
	extern int usbResetTick;
	tasklet_schedule(&gpio_tasklet_handle);
	mod_timer(&gpio_timer, jiffies + GPIO_POLL_DELAY);
	
	if (usbResetTick > 0) usbResetTick--;
	if (cycleDockPowerTimer > 0) cycleDockPowerTimer--;
}

void gpio_force_update(void)
{
	tasklet_schedule(&gpio_tasklet_handle);
}

static int gpio_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	int ret = 0;
	
	switch (cmd) 
	{
	case IOR_HWSTATUS:
		tasklet_disable(&gpio_tasklet_handle);
		ret = copy_to_user((void __user *) arg, &gpio_hw_status, sizeof gpio_hw_status) ? -EFAULT : 0;
		statusChanged = 0;
		tasklet_enable(&gpio_tasklet_handle);
		break;

	case IOR_DISK_ACCESS:
		tasklet_disable(&gpio_tasklet_handle);
		ret = copy_to_user((void __user *) arg, &diskAccessStarted, sizeof diskAccessStarted) ? -EFAULT: 0;
		diskAccessStarted = 0;
		tasklet_enable(&gpio_tasklet_handle);
		break;
	
	case IOR_BT_MODE:
		tasklet_disable(&gpio_tasklet_handle);
		ret = copy_from_user(&btModeState, (void __user *) arg, sizeof btModeState) ? -EFAULT: 0;
		tasklet_enable(&gpio_tasklet_handle);
		/* set BT mode */
		if (btModeState & 0x01)
			IO_Activate(BT_MODE);
		else
			IO_Deactivate(BT_MODE);
		/* issue BT reset */
		IO_Activate(BT_RESET);
		mdelay(8); /* wait at least 5ms according to BC4ROM datasheet */
		IO_Deactivate(BT_RESET);
		mdelay(8);
		IO_Activate(BT_RESET);
		mdelay(8);
		IO_Deactivate(BT_RESET);
		mdelay(100);
		break;

	case IOR_GET_BT_ERROR:
		ret = copy_to_user((void __user *) arg, &bt_failed, sizeof bt_failed) ? -EFAULT : 0;
		break;

	case IOR_SET_BT_ERROR:
		ret = copy_from_user(&bt_failed, (void __user *) arg, sizeof bt_failed) ? -EFAULT: 0;
		break;

	case IOW_RESET_ONOFF_STATE:
		tasklet_disable(&gpio_tasklet_handle);
		gpio_hw_status.u8InputStatus &= (~(ONOFF_MASK));
		tasklet_enable(&gpio_tasklet_handle);
		break;

	case IOW_ENABLE_DOCK_UART:
		{
			static unsigned int uartEnableState;
			tasklet_disable(&gpio_tasklet_handle);
			ret = copy_from_user(&uartEnableState, (void __user *) arg, sizeof uartEnableState) ? -EFAULT: 0;
			tasklet_enable(&gpio_tasklet_handle);
			if (uartEnableState) 
			{
				IO_SetFunction(TXD_DOCK);
				IO_SetFunction(RXD_DOCK);
			}
			else
			{
				IO_SetInput(TXD_DOCK);
				IO_SetInput(RXD_DOCK);
			}
		}
		break;
		
	case IOW_SET_FM_FREQUENCY:
			tasklet_disable(&gpio_tasklet_handle);
			ret = copy_from_user(&frequency, (void __user *) arg, sizeof frequency) ? -EFAULT: 0;
			tasklet_enable(&gpio_tasklet_handle);
		break;

	case IOW_CYCLE_DOCK_POWER:
			tasklet_disable(&gpio_tasklet_handle);
			cycleDockPowerTimer = 10; // Cycle power for 2 seconds
			tasklet_enable(&gpio_tasklet_handle);
		break;

  default:
		PK_WARN("Invalid ioctl command %u\n", cmd);
		ret = -EINVAL;
		break;
	}

	return ret;
}

static int gpio_open(struct inode *inode, struct file *file)
{
	return nonseekable_open(inode, file);
}

static int gpio_release(struct inode *inode, struct file *file)
{
	return 0;
}

static unsigned int gpio_poll(struct file *file, struct poll_table_struct *wait)
{
	unsigned int ret;

	poll_wait(file, &gpio_wait, wait);

	tasklet_disable(&gpio_tasklet_handle);
	ret = (statusChanged == 0) ? 0 : (POLLIN | POLLRDNORM);
	tasklet_enable(&gpio_tasklet_handle);

	return ret;
}

/* Kernel interface */
static struct file_operations gpio_fops = 
{
	.owner		= THIS_MODULE,
	.ioctl		= gpio_ioctl,
	.open		= gpio_open,
	.release	= gpio_release,
	.poll		= gpio_poll,
};

static void gpio_hw_init(void)
{
	IO_SetInput(ON_OFF);
	IO_SetInput(IGNITION);
	IO_SetInput(ACPWR);
	IO_SetInput(CHARGEFAULT);
	IO_SetInput(CHARGING);
	IO_SetInput(LIGHTS_DETECT);
	IO_SetInput(HEADPHONE_DETECT);
	IO_SetInput(EXTMIC_DETECT);
	IO_SetInput(LINEIN_DETECT);
	IO_SetInput(DOCK_SENSE);
	IO_SetInput(DOCK_DESK_SENSE);
	IO_SetInput(DOCK_MOTOR_SENSE);
	IO_SetInput(USB_HOST_DETECT);
	IO_SetInput(RXD_DOCK_INT);

	// Stop timers
	power_button_timer = -1;
	ignition_timer = -1;
	// Clear state from before suspend
	gpio_hw_status.u8InputStatus = 0;
	// Force update of transmitter if docked
	oldFrequency = 0;
	atomic_set(&uart_break_cnt, 0);
	// Force TMC/dock power to be enabled
	cycleDockPowerTimer = 0;

	// Get new state
	tasklet_schedule(&gpio_tasklet_handle);
	mod_timer(&gpio_timer, jiffies + GPIO_POLL_DELAY);
}

static void gpio_hw_exit(void)
{
	del_timer_sync(&gpio_timer);
}

static int gpio_probe(struct device *dev)
{
	int ret;

	PK_DBG("Initializing wait queue\n");
	init_waitqueue_head(&gpio_wait);

	gpio_hw_init();

	if (IO_HaveUsbBusPowered()) 
	{
		ret = buspower_probe(dev);
		if (ret < 0) 
		{
			printk("GPIO driver: buspower sub-driver failed (%d)\n", ret);
			return ret;
		}
	}

	PK_DBG("Registering chardev\n");
	ret = register_chrdev(GPIO_MAJOR, GPIO_DEVNAME, &gpio_fops);
	if (ret < 0) 
	{
		printk("Unable to register chardev on major=%d (%d)\n", GPIO_MAJOR, ret);
		return ret;
	}
		
	printk("Done\n");
	return 0;
}

static int gpio_remove(struct device *dev)
{
	PK_DBG("Unregistering chardev\n");
	unregister_chrdev(GPIO_MAJOR, GPIO_DEVNAME);

	if (IO_HaveUsbBusPowered()) 
	{
		buspower_remove(dev);
	}

	gpio_hw_exit();

	PK_DBG("Done\n");
	return 0;
}

static void gpio_shutdown(struct device *dev)
{
	PK_DBG("Shutting down\n");
	gpio_hw_exit();
}

#ifdef CONFIG_PM

static int gpio_suspend(struct device *dev, u32 state, u32 level)
{
	PK_DBG("dev = %p, state = %u, level = %u\n", dev, state, level);
	if (level == SUSPEND_POWER_DOWN) 
	{
		gpio_hw_exit();
	}
	return 0;
}

static int gpio_resume(struct device *dev, u32 level)
{
	PK_DBG("dev = %p, level = %u\n", dev, level);
	if (level == RESUME_POWER_ON) 
	{
		gpio_hw_init();
	}
	return 0;
}

#else /* CONFIG_PM */
#define gpio_suspend NULL
#define gpio_resume  NULL
#endif /* CONFIG_PM */

static struct device_driver gpio_driver = 
{
	.name		= "tomtomgo-gpio",
	.bus		= &platform_bus_type,
	.probe		= gpio_probe,
	.remove		= gpio_remove,
	.shutdown	= gpio_shutdown,
	.suspend	= gpio_suspend,
	.resume		= gpio_resume,
};

static int __init gpio_mod_init(void)
{
	int ret;

	printk(KERN_INFO "TomTom GO GPIO Driver, (C) 2004,2005 TomTom BV\n");
	PK_DBG("Registering driver\n");
	ret = driver_register(&gpio_driver);
	if (ret) 
	{
		PK_ERR("Unable to register driver (%d)\n", ret);
		return ret;
	}
	PK_DBG("Done\n");
	return 0;
}

static void __exit gpio_mod_exit(void)
{
	PK_DBG("Unregistering driver\n");
	driver_unregister(&gpio_driver);
	PK_DBG("Done\n");
}

module_init(gpio_mod_init);
module_exit(gpio_mod_exit);

MODULE_AUTHOR("Jeroen Taverne <jeroen.taverne@tomtom.com> and Dimitry Andric <dimitry.andric@tomtom.com> and Koen Martens <kmartens@sonologic.nl>");
MODULE_DESCRIPTION("TomTom GO GPIO Driver");
MODULE_LICENSE("GPL");

/* EOF */
