/*
 *  linux/drivers/mmc/s3c2410mci.h - Samsung S3C2410 SDI Interface driver
 *
 *  Copyright (C) 2004-2005 Thomas Kleffel, All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/dma-mapping.h>
#include <linux/mmc/host.h>
#include <linux/mmc/protocol.h>

#include <asm/dma.h>
#include <asm/dma-mapping.h>

#include <asm/io.h>
#include <asm/hardware/clock.h>

#include <asm/arch/regs-sdi.h>
#include <asm/arch/regs-gpio.h>

#include <barcelona/gopins.h>

#include "s3c2410mci.h"

#define DRIVER_NAME "s3c24XX-mci"
#define PFX DRIVER_NAME ": "

#define NULLDBG(x...)	do { } while (0)
#define KERNDBG(x...)	printk(KERN_DEBUG PFX x)
#define KERNINF(x...)	printk(KERN_INFO PFX x)
#define KERNERR(x...)	printk(KERN_ERR PFX x)

#define ERR(x...)	KERNERR("ERR " x)
#define DBG(x...)	NULLDBG("DBG " x)
#define INFO(x...)	KERNINF("INF " x)
#define INTDBG(x...)	NULLDBG("INT " x)
#define SGDBG(x...)	NULLDBG("SGD " x)
#define DMADBG(x...)	NULLDBG("DMA " x)
#define PIODBG(x...)	NULLDBG("PIO " x)
#define FAILDBG(x...)	KERNERR("FRQ " x)
#define CONFDBG(x...)	KERNINF("CFG " x)

#define RESSIZE(ressource) (((ressource)->end - (ressource)->start)+1)

typedef enum {
	DMAP_READ,
	DMAP_WRITE,
} eDMAPurpose_t;

static struct s3c2410_dma_client s3c2410sdi_dma_client = {
	.name		= "s3c2410-sdi",
};

static void finalize_request(struct s3c2410sdi_host *host);
static void s3c2410sdi_send_request(struct mmc_host *mmc);
static void s3c2410sdi_reset(struct s3c2410sdi_host *host);
static void s3c2410sdi_write_ios_settings( struct s3c2410sdi_host *host );

static inline void debug_dump_registers(struct s3c2410sdi_host *host, char *prefix) {
	u32 con, pre, cmdarg, cmdcon, cmdsta, r0, r1, r2, r3, timer, bsize;
	u32 datcon, datcnt, datsta, fsta, imask;

	con 	= readl(host->base + S3C2410_SDICON);
	pre 	= readl(host->base + S3C2410_SDIPRE);
	cmdarg 	= readl(host->base + S3C2410_SDICMDARG);
	cmdcon 	= readl(host->base + S3C2410_SDICMDCON);
	cmdsta 	= readl(host->base + S3C2410_SDICMDSTAT);
	r0 	= readl(host->base + S3C2410_SDIRSP0);
	r1 	= readl(host->base + S3C2410_SDIRSP1);
	r2 	= readl(host->base + S3C2410_SDIRSP2);
	r3 	= readl(host->base + S3C2410_SDIRSP3);
	timer 	= readl(host->base + S3C2410_SDITIMER);
	bsize 	= readl(host->base + S3C2410_SDIBSIZE);
	datcon 	= readl(host->base + S3C2410_SDIDCON);
	datcnt 	= readl(host->base + S3C2410_SDIDCNT);
	datsta 	= readl(host->base + S3C2410_SDIDSTA);
	fsta 	= readl(host->base + S3C2410_SDIFSTA);
	imask   = readl(host->base + host->sdiimsk);

	DBG("%s  CON:[%08x]  PRE:[%08x]  TMR:[%08x]\n", prefix, con, pre, timer);
	DBG("%s CCON:[%08x] CARG:[%08x] CSTA:[%08x]\n", prefix, cmdcon, cmdarg, cmdsta);
	DBG("%s DCON:[%08x] FSTA:[%08x] DSTA:[%08x] DCNT:[%08x]\n", prefix, datcon, fsta, datsta, datcnt);
	DBG("%s   R0:[%08x]   R1:[%08x]   R2:[%08x]   R3:[%08x]\n", prefix, r0, r1, r2, r3);
}

static inline char *cmd2str(int cmd) {
	switch(cmd) {
		case  0: return "GO_IDLE_STATE";
		case  1: return "ALL_SEND_OCR";
		case  2: return "ALL_SEND_CID";
		case  3: return "ALL_SEND_RELATIVE_ADD";
		case  6: return "ACMD: SD_SET_BUSWIDTH";
		case  7: return "SEL_DESEL_CARD";
		case  9: return "SEND_CSD";
		case 10: return "SEND_CID";
		case 11: return "READ_UNTIL_STOP";
		case 12: return "STOP_TRANSMISSION";
		case 13: return "SEND_STATUS";
		case 15: return "GO_INACTIVE_STATE";
		case 16: return "SET_BLOCKLEN";
		case 17: return "READ_SINGLE_BLOCK";
		case 18: return "READ_MULTIPLE_BLOCK";
		case 24: return "WRITE_SINGLE_BLOCK";
		case 25: return "WRITE_MULTIPLE_BLOCK";
		case 41: return "ACMD: SD_APP_OP_COND";
		case 51: return "ACMD: SD_APP_SEND_SCR";
		case 55: return "APP_CMD";
		default: return "UNKNOWN";
	}
}

static inline char *err2str(int err) {
	switch(err) {
		case 0: return "OK";
		case 1: return "TIMEOUT";
		case 2: return "BADCRC";
		case 3: return "FIFO";
		case 4: return "FAILED";
		case 5: return "INVALID";
		case 6: return "BUSY";
		case 7: return "DMA";
		case 8: return "CANCELED";
		default: return "UNKNOWN";
	}
}

static inline void prepare_dbgmsg(struct s3c2410sdi_host *host, struct mmc_command *cmd, int stop)
{
 	snprintf(host->dbgmsg_cmd, 300,
		"#[%u]%s op:[%s(%i)] arg:[0x%08x] flags:[0x08%x] retries:[%u]",
		host->ccnt, (stop?" STOP":""), cmd2str(cmd->opcode),
		cmd->opcode, cmd->arg, cmd->flags, cmd->retries);

	if(cmd->data)
		snprintf(host->dbgmsg_dat, 300,
			"#[%u] bsize:[%u] blocks:[%u] bytes:[%u]	",
			host->dcnt, (1 << cmd->data->blksz_bits) , cmd->data->blocks,
			cmd->data->blocks * (1 << cmd->data->blksz_bits));
	else
		host->dbgmsg_dat[0] = '\0';
}

#define DBGCMD(x...) 	if(fail) {FAILDBG(x);} else {DBG(x);}
static void debug_command(struct s3c2410sdi_host *host, struct mmc_command *cmd, int fail)
{
	if(!cmd) return;

	if(cmd->error == MMC_ERR_NONE) {
		DBGCMD("CMD[OK] %s R0:[0x%08x]\n", host->dbgmsg_cmd, cmd->resp[0]);
	} else {
		DBGCMD("CMD[%s] %s Status:[%s]\n", err2str(cmd->error), host->dbgmsg_cmd, host->status);
	}

	if(!cmd->data) return;

	if(cmd->data->error == MMC_ERR_NONE) {
		DBGCMD("DAT[%s] %s\n", err2str(cmd->data->error), host->dbgmsg_dat);
	} else {
		DBGCMD("DAT[%s] %s DCNT:[0x%08x]\n", err2str(cmd->data->error), host->dbgmsg_dat,
			readl(host->base + S3C2410_SDIDCNT));
	}
}

static inline u32 enable_imask(struct s3c2410sdi_host *host, u32 imask) {
	u32 newmask;

	newmask = readl(host->base + host->sdiimsk);
	newmask|= imask;

	writel(newmask, host->base + host->sdiimsk);

	return newmask;
}

static inline u32 disable_imask(struct s3c2410sdi_host *host, u32 imask) {
	u32 newmask;

	newmask = readl(host->base + host->sdiimsk);
	newmask&= ~imask;

	writel(newmask, host->base + host->sdiimsk);

	return newmask;
}

static inline void clear_imask(struct s3c2410sdi_host *host) {
	writel(0, host->base + host->sdiimsk);
}

static inline int get_data_buffer(struct s3c2410sdi_host *host, volatile u32 *words, volatile u32 **pointer)
{
	struct scatterlist *sg;

	if(host->pio_active == XFER_NONE) return -EINVAL;
	if(!host->mrq) return -EINVAL;
	if(!host->mrq->data) return -EINVAL;

	if(host->pio_sgptr >= host->mrq->data->sg_len) {
		SGDBG("get_data_buffer(): no more buffers (%i/%i)\n",
		      host->pio_sgptr, host->mrq->data->sg_len);
		return -EBUSY;
	}
	sg = &host->mrq->data->sg[host->pio_sgptr];

	*words	= sg->length >> 2;
	*pointer= page_address(sg->page) + sg->offset;

	host->pio_sgptr++;

	SGDBG("get_data_buffer(): new buffer (%i/%i)\n",
	      host->pio_sgptr, host->mrq->data->sg_len);

	return 0;
}

#define FIFO_FILL(host) ((readl(host->base + S3C2410_SDIFSTA) & S3C2410_SDIFSTA_COUNTMASK) >> 2)
#define FIFO_FREE(host) ((63 - (readl(host->base + S3C2410_SDIFSTA) & S3C2410_SDIFSTA_COUNTMASK)) >> 2)

static inline void do_pio_read(struct s3c2410sdi_host *host)
{
	int res;
	u32 fifo;
	void __iomem *from_ptr;

	//FIX2410: write real prescaler to host, it might be set slow to fix 2410
	writel(host->prescaler, host->base + S3C2410_SDIPRE);

	from_ptr = host->base + host->sdidata;

	while((fifo = FIFO_FILL(host))) {
		if(!host->pio_words) {
			res = get_data_buffer(host, &host->pio_words, &host->pio_ptr);
			if(res) {
				host->pio_active = XFER_NONE;
				host->complete_what = COMPLETION_FINALIZE;

				PIODBG("pio_read(): complete (no more buffers).\n");
				return;
				//TODO: handle no data condition
			}

			PIODBG("pio_read(): new target: [%i]@[%p]\n",
			       host->pio_words, host->pio_ptr);
		}

		PIODBG("pio_read(): fifo:[%02i] buffer:[%03i] dcnt:[%08X]\n",
		       fifo, host->pio_words, readl(host->base + S3C2410_SDIDCNT));

		if(fifo > host->pio_words)
			fifo = host->pio_words;

		host->pio_words-= fifo;
		host->pio_count+= fifo;

		while(fifo--) {
			*(host->pio_ptr++) = readl(from_ptr);
		}
	}

	if(!host->pio_words) {
		res = get_data_buffer(host, &host->pio_words, &host->pio_ptr);
		if(res) {
			PIODBG("pio_read(): complete (no more buffers).\n");
			host->pio_active = XFER_NONE;
			host->complete_what = COMPLETION_FINALIZE;

			return;
		}
	}

	enable_imask(host, S3C2410_SDIIMSK_RXFIFOHALF | S3C2410_SDIIMSK_RXFIFOLAST);
}

static inline void do_pio_write(struct s3c2410sdi_host *host)
{
	int res;
	u32 fifo;

	void __iomem *to_ptr;

	to_ptr = host->base + host->sdidata;

	while((fifo = FIFO_FREE(host))) {
		if(!host->pio_words) {
			res = get_data_buffer(host, &host->pio_words, &host->pio_ptr);
			if(res) {
				PIODBG("pio_write(): complete (no more buffers).\n");
				host->pio_active = XFER_NONE;

				return;
			}

			PIODBG("pio_write(): new source: [%i]@[%p]\n",
			       host->pio_words, host->pio_ptr);

		}

		if(fifo > host->pio_words)
			fifo = host->pio_words;

		host->pio_words-= fifo;
		host->pio_count+= fifo;

		while(fifo--) {
			writel(*(host->pio_ptr++), to_ptr);
		}
	}

	enable_imask(host, S3C2410_SDIIMSK_TXFIFOHALF);
}

static void pio_tasklet(unsigned long data) {
	struct s3c2410sdi_host *host = (struct s3c2410sdi_host *) data;

	disable_irq(host->irq);

	if(host->pio_active == XFER_WRITE)	do_pio_write(host);
	if(host->pio_active == XFER_READ)	do_pio_read(host);

	if(host->complete_what == COMPLETION_FINALIZE) {
		clear_imask(host);
		if(host->pio_active != XFER_NONE) {
			ERR("unfinished %s - pio_count:[%u] pio_words:[%u]\n",
				(host->pio_active == XFER_READ)?"read":"write", host->pio_count, host->pio_words);
			debug_command(host, host->cmd_is_stop?host->mrq->stop:host->mrq->cmd, 1);

		}

		finalize_request(host);
	} else
		enable_irq(host->irq);
}

/*
 * ISR for SDI Interface IRQ
 * Communication between driver and ISR works as follows:
 *   host->mrq 			points to current request
 *   host->complete_what	tells the ISR when the request is considered done
 *     COMPLETION_CMDSENT	  when the command was sent
 *     COMPLETION_RSPFIN          when a response was received
 *     COMPLETION_XFERFINISH	  when the data transfer is finished
 *     COMPLETION_XFERFINISH_RSPFIN both of the above.
 *   host->complete_request	is the completion-object the driver waits for
 *
 * 1) Driver sets up host->mrq and host->complete_what
 * 2) Driver prepares the transfer
 * 3) Driver enables interrupts
 * 4) Driver starts transfer
 * 5) Driver waits for host->complete_rquest
 * 6) ISR checks for request status (errors and success)
 * 6) ISR sets host->mrq->cmd->error and host->mrq->data->error
 * 7) ISR completes host->complete_request
 * 8) ISR disables interrupts
 * 9) Driver wakes up and takes care of the request
 *
 * Note: "->error"-fields are expected to be set to 0 before the request
 *       was issued by mmc.c - therefore they are only set, when an error
 *       contition comes up
 */

static irqreturn_t s3c2410sdi_irq(int irq, void *dev_id, struct pt_regs *regs) {
	struct s3c2410sdi_host *host;
	struct mmc_command *cmd;
	u32 sdi_csta, sdi_dsta, sdi_fsta, sdi_dcnt, sdi_imsk;
	u32 sdi_cclear, sdi_dclear;
	unsigned long iflags;

	host = (struct s3c2410sdi_host *)dev_id;

	//Check for things not supposed to happen
	if(!host) return IRQ_HANDLED;

	spin_lock_irqsave(&host->complete_lock, iflags);

	sdi_csta 	= readl(host->base + S3C2410_SDICMDSTAT);
	sdi_dsta 	= readl(host->base + S3C2410_SDIDSTA);
	sdi_dcnt 	= readl(host->base + S3C2410_SDIDCNT);
	sdi_fsta 	= readl(host->base + S3C2410_SDIFSTA);
	sdi_imsk	= readl(host->base + host->sdiimsk);
	sdi_cclear	= 0;
	sdi_dclear	= 0;

	if((host->complete_what == COMPLETION_NONE) || (host->complete_what == COMPLETION_FINALIZE)) {
		host->status = "nothing to complete";
		clear_imask(host);
		goto irq_out;
	}

	if(!host->mrq) {
		host->status = "no active mrq";
		clear_imask(host);
		goto irq_out;
	}

	cmd = host->cmd_is_stop?host->mrq->stop:host->mrq->cmd;

	if(!cmd) {
		host->status = "no active cmd";
		clear_imask(host);
		goto irq_out;
	}

	if(!host->dodma) {
		if((host->pio_active == XFER_WRITE) && (sdi_fsta & S3C2410_SDIFSTA_TFDET)) {
			disable_imask(host, S3C2410_SDIIMSK_TXFIFOHALF);
			tasklet_schedule(&host->pio_tasklet);
			host->status = "pio tx";
		}

		if((host->pio_active == XFER_READ) && (sdi_fsta & S3C2410_SDIFSTA_RFDET)) {
			disable_imask(host, S3C2410_SDIIMSK_RXFIFOHALF | S3C2410_SDIIMSK_RXFIFOLAST);
			tasklet_schedule(&host->pio_tasklet);
			host->status = "pio rx";
		}
	}

	if(sdi_csta & S3C2410_SDICMDSTAT_CMDTIMEOUT) {
		cmd->error = MMC_ERR_TIMEOUT;
		host->status = "error: command timeout";
		goto fail_transfer;
	}

	if(sdi_csta & S3C2410_SDICMDSTAT_CMDSENT) {
		if(host->complete_what == COMPLETION_CMDSENT) {
			host->status = "ok: command sent";
			goto close_transfer;
		}

		sdi_cclear |= S3C2410_SDICMDSTAT_CMDSENT;
	}

	if(sdi_csta & S3C2410_SDICMDSTAT_CRCFAIL) {
		if(cmd->flags & MMC_RSP_CRC) {
			cmd->error = MMC_ERR_BADCRC;
			host->status = "error: bad command crc";
			goto fail_transfer;
		}

		sdi_cclear |= S3C2410_SDICMDSTAT_CRCFAIL;
	}

	if(sdi_csta & S3C2410_SDICMDSTAT_RSPFIN) {
		if(host->complete_what == COMPLETION_RSPFIN) {
			host->status = "ok: command response received";
			goto close_transfer;
		}

		if(host->complete_what == COMPLETION_XFERFINISH_RSPFIN) {
			host->complete_what = COMPLETION_XFERFINISH;
		}

		sdi_cclear |= S3C2410_SDICMDSTAT_RSPFIN;
	}

	/* errors handled after this point are only relevant
	   when a data transfer is in progress */

	if(!cmd->data) goto clear_status_bits;

	/* Check for FIFO failure */
	if(host->is2440) {
		if(sdi_fsta & S3C2440_SDIFSTA_FIFOFAIL) {
			host->mrq->data->error = MMC_ERR_FIFO;
			host->status = "error: 2440 fifo failure";
			goto fail_transfer;
		}
	} else {
		if(sdi_dsta & S3C2410_SDIDSTA_FIFOFAIL) {
			cmd->data->error = MMC_ERR_FIFO;
			host->status = "error: 2410 fifo failure";
			goto fail_transfer;
		}
	}

	if(sdi_dsta & S3C2410_SDIDSTA_RXCRCFAIL) {
		cmd->data->error = MMC_ERR_BADCRC;
		host->status = "error: bad data crc (outgoing)";
		goto fail_transfer;
	}

	if(sdi_dsta & S3C2410_SDIDSTA_CRCFAIL) {
		cmd->data->error = MMC_ERR_BADCRC;
		host->status = "error: bad data crc (incoming)";
		goto fail_transfer;
	}

	if(sdi_dsta & S3C2410_SDIDSTA_DATATIMEOUT) {
		cmd->data->error = MMC_ERR_TIMEOUT;
		host->status = "error: data timeout";
		goto fail_transfer;
	}

	if(sdi_dsta & S3C2410_SDIDSTA_XFERFINISH) {
		if(host->complete_what == COMPLETION_XFERFINISH) {
			host->status = "ok: data transfer completed";
			goto close_transfer;
		}

		if(host->complete_what == COMPLETION_XFERFINISH_RSPFIN) {
			host->complete_what = COMPLETION_RSPFIN;
		}

		sdi_dclear |= S3C2410_SDIDSTA_XFERFINISH;
	}

clear_status_bits:
	writel(sdi_cclear, host->base + S3C2410_SDICMDSTAT);
	writel(sdi_dclear, host->base + S3C2410_SDIDSTA);

	goto irq_out;

fail_transfer:
	host->pio_active = XFER_NONE;

close_transfer:
	host->complete_what = COMPLETION_FINALIZE;

	clear_imask(host);
	tasklet_schedule(&host->pio_tasklet);

	goto irq_out;

irq_out:
	INTDBG("csta:[0x%08x] dsta:[0x%08x] fsta:[0x%08x] dcnt:[0x%08x] status:[%s].\n", sdi_csta, sdi_dsta, sdi_fsta, sdi_dcnt, host->status);

	spin_unlock_irqrestore(&host->complete_lock, iflags);
	return IRQ_HANDLED;

}

/*
 * ISR for the CardDetect Pin
*/

static irqreturn_t s3c2410sdi_irq_cd(int irq, void *dev_id, struct pt_regs *regs)
{
	struct s3c2410sdi_host *host = (struct s3c2410sdi_host *)dev_id;

	INTDBG("CardDetect IRQ.\n");

	mmc_detect_change(host->mmc);

	return IRQ_HANDLED;
}

void s3c2410sdi_dma_done_callback(s3c2410_dma_chan_t *dma_ch, void *buf_id,
	int size, s3c2410_dma_buffresult_t result)
{	unsigned long iflags;
	u32 sdi_csta, sdi_dsta, sdi_fsta, sdi_dcnt;
	struct s3c2410sdi_host *host = (struct s3c2410sdi_host *)buf_id;

	sdi_csta 	= readl(host->base + S3C2410_SDICMDSTAT);
	sdi_dsta 	= readl(host->base + S3C2410_SDIDSTA);
	sdi_fsta 	= readl(host->base + S3C2410_SDIFSTA);
	sdi_dcnt 	= readl(host->base + S3C2410_SDIDCNT);

	if(!host->mrq) return;
	if(!host->mrq->data) return;
	if(!host->dmatogo) return;

	spin_lock_irqsave(&host->complete_lock, iflags);


	if(result != S3C2410_RES_OK) {
		FAILDBG("DMA FAILED: csta=0x%08x dsta=0x%08x fsta=0x%08x dcnt:0x%08x result:0x%08x togo:%u\n", sdi_csta, sdi_dsta, sdi_fsta, sdi_dcnt, result, host->dmatogo);
		goto fail_request;
	}

	host->dmatogo--;
	if(host->dmatogo) {
		DMADBG("DMA DONE %i (%i) togo DSTA:[%08x] DCNT:[%08x]\n", host->dmatogo, size, sdi_dsta, sdi_dcnt);
		goto out;
	}

	DMADBG("DMA FINISHED (%i)  DSTA:[%08x] DCNT:[%08x]\n", size, sdi_dsta, sdi_dcnt);
	host->complete_what = COMPLETION_FINALIZE;

out:
	tasklet_schedule(&host->pio_tasklet);
	spin_unlock_irqrestore(&host->complete_lock, iflags);
	return;


fail_request:
	host->mrq->data->error = MMC_ERR_DMA;
	host->complete_what = COMPLETION_FINALIZE;
	writel(0, host->base + host->sdiimsk);
	goto out;

}

static void finalize_request(struct s3c2410sdi_host *host) {
	struct mmc_request *mrq;
	struct mmc_command *cmd;
	int debug_as_failure;

	if(host->complete_what != COMPLETION_FINALIZE) return;

	mrq = host->mrq;
	if(!mrq) return;

	cmd = host->cmd_is_stop?mrq->stop:mrq->cmd;
	if(cmd->data && (cmd->error == MMC_ERR_NONE) &&
		  (cmd->data->error == MMC_ERR_NONE)) {

		if(host->dodma && (!host->dma_complete)) {
			DBG("DMA Missing!\n");
			s3c2410_dma_debug_channel(host->dma);
			return;
		}
	}

	// Read response
	cmd->resp[0] = readl(host->base + S3C2410_SDIRSP0);
	cmd->resp[1] = readl(host->base + S3C2410_SDIRSP1);
	cmd->resp[2] = readl(host->base + S3C2410_SDIRSP2);
	cmd->resp[3] = readl(host->base + S3C2410_SDIRSP3);

	// FIX2410: reset clock speed, as it could still be set low for 2410
	writel(host->prescaler, host->base + S3C2410_SDIPRE);

	debug_as_failure = 0;
	if(cmd->error) debug_as_failure = 1;
	if(cmd->data && cmd->data->error) debug_as_failure = 1;
	if(cmd->flags & MMC_RSP_MAYFAIL) debug_as_failure = 0;

	debug_command(host, cmd, debug_as_failure);

	//Cleanup controller
	writel(0, host->base + S3C2410_SDICMDARG);
	writel(0, host->base + S3C2410_SDIDCON);
	writel(0, host->base + S3C2410_SDICMDCON);
	writel(0, host->base + host->sdiimsk);

	if(cmd->data && cmd->error) cmd->data->error = cmd->error;

	if(cmd->data && cmd->data->stop && (!host->cmd_is_stop)) {
		host->cmd_is_stop = 1;
		s3c2410sdi_send_request(host->mmc);
		return;
	}

	// If we have no data transfer we are finished here
	if (!mrq->data) goto request_done;

	// Calulate the amout of bytes transfer, but only if there was
	// no error
	if(mrq->data->error == MMC_ERR_NONE) {
		mrq->data->bytes_xfered = (mrq->data->blocks << mrq->data->blksz_bits);
	} else {
		mrq->data->bytes_xfered = 0;
	}

	// If we had an error while transfering data we flush the
	// DMA channel and the fifo to clear out any garbage
	if(mrq->data->error != MMC_ERR_NONE) {
		if(host->dodma) {
			s3c2410_dma_ctrl(host->dma, S3C2410_DMAOP_FLUSH);
		}

		if(host->is2440) {
			//Clear failure register and reset fifo
			writel(S3C2440_SDIFSTA_FIFORESET |
			       S3C2440_SDIFSTA_FIFOFAIL,
			       host->base + S3C2410_SDIFSTA);
		} else {
			u32 sdi_con;

			//reset fifo
			sdi_con = readl(host->base + S3C2410_SDICON);
			sdi_con|= S3C2410_SDICON_FIFORESET;

			writel(sdi_con, host->base + S3C2410_SDICON);
		}
	}

request_done:
	host->complete_what = COMPLETION_NONE;
	host->mrq = NULL;
	mmc_request_done(host->mmc, mrq);
}


void s3c2410sdi_dma_setup(struct s3c2410sdi_host *host, eDMAPurpose_t purpose) {
	static int setup_ok = 0;
	static eDMAPurpose_t lp = -1;
	s3c2410_dmasrc_t source;

	if(lp == purpose) return;
	lp = purpose;

	switch(purpose) {
		case DMAP_READ:
			source  = S3C2410_DMASRC_HW;
			break;

		case DMAP_WRITE:
			source  = S3C2410_DMASRC_MEM;
			break;

		default:
			return;
	}

	s3c2410_dma_devconfig(host->dma, source, 3, host->mem->start + host->sdidata, 2);
	if(!setup_ok) {
		s3c2410_dma_config(host->dma, 4, (1<<23));
		s3c2410_dma_set_buffdone_fn(host->dma, s3c2410sdi_dma_done_callback);
		s3c2410_dma_setflags(host->dma, S3C2410_DMAF_AUTOSTART);
		setup_ok = 1;
	}
}

static void s3c2440sdi_send_command(struct s3c2410sdi_host *host, struct mmc_command *cmd) {
	u32 ccon, imsk;

	imsk  = S3C2410_SDIIMSK_CRCSTATUS | S3C2410_SDIIMSK_CMDTIMEOUT |
		S3C2410_SDIIMSK_RESPONSEND | S3C2410_SDIIMSK_CMDSENT |
		S3C2410_SDIIMSK_RESPONSECRC;

	enable_imask(host, imsk);

	if(cmd->data)
		host->complete_what = COMPLETION_XFERFINISH_RSPFIN;
	else if (cmd->flags & MMC_RSP_MASK)
		host->complete_what = COMPLETION_RSPFIN;
	else
		host->complete_what = COMPLETION_CMDSENT;

	writel(cmd->arg, host->base + S3C2410_SDICMDARG);

	ccon = cmd->opcode & S3C2410_SDICMDCON_INDEX;
	ccon|= S3C2410_SDICMDCON_SENDERHOST | S3C2410_SDICMDCON_CMDSTART;

	if(cmd->flags & MMC_RSP_MASK)
		ccon |= S3C2410_SDICMDCON_WAITRSP;

	if(cmd->flags & MMC_RSP_LONG)
		ccon|= S3C2410_SDICMDCON_LONGRSP;
	/*
	DBG("Sending Command: CCON:[%08X] DCON:[%08X] BSIZE:[%08X] IMASK:[%08X] DCNT:[%08X]\n",
	       ccon, readl(host->base + S3C2410_SDIDCON), readl(host->base + S3C2410_SDIBSIZE),
	       readl(host->base + host->sdiimsk),
	       readl(host->base + S3C2410_SDIDCNT));
	*/
	writel(ccon, host->base + S3C2410_SDICMDCON);
}

static int s3c2440sdi_setup_data(struct s3c2410sdi_host *host, struct mmc_data *data) {
	u32 dcon, imsk, stoptries=3;

	/* write DCON register */

	if(!data) {
		writel(0, host->base + S3C2410_SDIDCON);
		return 0;
	}

	while(readl(host->base + S3C2410_SDIDSTA) & (S3C2410_SDIDSTA_TXDATAON | S3C2410_SDIDSTA_RXDATAON)) {
		ERR("sdi_setup_data() transfer in progress. Trying to stop...\n");

		writel(0, host->base + S3C2410_SDIDCON);
		s3c2410sdi_reset(host);

		/* Obtain a write lock when writing to hardware	*/
		write_lock_bh( &host->ios_lock );
		{
			/* Write ios in cache to hardware	*/
			s3c2410sdi_write_ios_settings( host );
		}
		write_unlock_bh( &host->ios_lock );

		if(0 == (stoptries--)) {
			debug_dump_registers(host, "Data recovery:");

			//TODO: get MCI-Controller out of data transfer mode, somehow....
			return -EINVAL;
		}
	}

	dcon  = data->blocks & S3C2410_SDIDCON_BLKNUM_MASK;

	if(host->dodma)
		dcon |= S3C2410_SDIDCON_DMAEN;

	if(data->flags & MMC_DATA_WIDE)
		dcon |= S3C2410_SDIDCON_WIDEBUS;

	if(!(data->flags & MMC_DATA_STREAM))
		dcon |= S3C2410_SDIDCON_BLOCKMODE;

	if(data->flags & MMC_DATA_WRITE) {
		dcon |= S3C2410_SDIDCON_TXAFTERRESP;
		dcon |= S3C2410_SDIDCON_XFER_TXSTART;
	}

	if(data->flags & MMC_DATA_READ) {
		dcon |= S3C2410_SDIDCON_RXAFTERCMD;
		dcon |= S3C2410_SDIDCON_XFER_RXSTART;
	}

	if(host->is2440) {
		dcon |= S3C2440_SDIDCON_DS_WORD;
		dcon |= S3C2440_SDIDCON_DATSTART;
	}

	writel(dcon, host->base + S3C2410_SDIDCON);

	/* write BSIZE register */

	writel((1 << data->blksz_bits), host->base + S3C2410_SDIBSIZE);

	/* add to IMASK register */
	imsk =	S3C2410_SDIIMSK_FIFOFAIL | S3C2410_SDIIMSK_DATACRC |
		S3C2410_SDIIMSK_DATATIMEOUT | S3C2410_SDIIMSK_DATAFINISH;

	enable_imask(host, imsk);

	/* write TIMER register */

	if(host->is2440) {
		writel(0x007FFFFF, host->base + S3C2410_SDITIMER);
	} else {
		writel(0x0000FFFF, host->base + S3C2410_SDITIMER);

		//FIX2410: set slow clock to prevent timeouts on read
		if(data->flags & MMC_DATA_READ) {
			writel(0xFF, host->base + S3C2410_SDIPRE);
		}
	}

	//debug_dump_registers(host, "Data setup:");

	return 0;
}

static int s3c2440sdi_prepare_pio(struct s3c2410sdi_host *host, struct mmc_data *data)
{
	int rw = (data->flags & MMC_DATA_WRITE)?1:0;
	if(rw != ((data->flags & MMC_DATA_READ)?0:1)) return -EINVAL;

	host->pio_sgptr = 0;
	host->pio_words = 0;
	host->pio_count = 0;
	host->pio_active = rw?XFER_WRITE:XFER_READ;

	if(rw) {
		do_pio_write(host);
		enable_imask(host, S3C2410_SDIIMSK_TXFIFOHALF);
	} else {
		enable_imask(host, S3C2410_SDIIMSK_RXFIFOHALF | S3C2410_SDIIMSK_RXFIFOLAST);
	}

	return 0;
}

static int s3c2440sdi_prepare_dma(struct s3c2410sdi_host *host, struct mmc_data *data)
{
	int dma_len, i;

	int rw = (data->flags & MMC_DATA_WRITE)?1:0;
	if(rw != ((data->flags & MMC_DATA_READ)?0:1)) return -EINVAL;

	s3c2410sdi_dma_setup(host, rw?DMAP_WRITE:DMAP_READ);
	s3c2410_dma_ctrl(host->dma, S3C2410_DMAOP_FLUSH);

	dma_len = dma_map_sg(mmc_dev(host->mmc), data->sg, data->sg_len,
				(rw)?DMA_TO_DEVICE:DMA_FROM_DEVICE);


	if(dma_len == 0) return -ENOMEM;

	host->dma_complete = 0;
	host->dmatogo = dma_len;

	for (i = 0; i < dma_len; i++) {
		int res;
		DBG("dma_enqueue %i:[%u]@[%u]\n", i, sg_dma_address(&data->sg[i]), sg_dma_len(&data->sg[i]));
		res = s3c2410_dma_enqueue(host->dma, (void *) host,
				sg_dma_address(&data->sg[i]), sg_dma_len(&data->sg[i]));

		if(res) {
			s3c2410_dma_ctrl(host->dma, S3C2410_DMAOP_FLUSH);
			return -EBUSY;
		}
 	}

	//s3c2410_dma_ctrl(host->dma, S3C2410_DMAOP_START);

	return 0;
}

static void s3c2410sdi_send_request(struct mmc_host *mmc) {
	struct s3c2410sdi_host *host = mmc_priv(mmc);
	struct mmc_request *mrq = host->mrq;
	struct mmc_command *cmd = host->cmd_is_stop?mrq->stop:mrq->cmd;
	unsigned char power_mode;

	/* Increment request counters */
	host->ccnt++;
	if(cmd->data) host->dcnt++;

	prepare_dbgmsg(host, cmd, host->cmd_is_stop);

	/* sample the current SDI power state	*/
	read_lock_bh( &host->ios_lock );
	power_mode = host->ios.power_mode;
	read_unlock_bh( &host->ios_lock );

	/* If we are not powered on, there is no use in issuing a request	*/
	/* so we fail is straight away to prevent a deadlock [FIX: GO-2174]	*/
	if( power_mode != MMC_POWER_ON ){
		cmd->error = MMC_ERR_FAILED;
		if( cmd->data)
			cmd->data->error = MMC_ERR_FAILED;
		mmc_request_done(mmc, mrq);
		return;
	}

	//Clear command, data and fifo status registers
	//Fifo clear only necessary on 2440, but doesn't hurt on 2410
	writel(0xFFFFFFFF, host->base + S3C2410_SDICMDSTAT);
	writel(0xFFFFFFFF, host->base + S3C2410_SDIDSTA);
	writel(0xFFFFFFFF, host->base + S3C2410_SDIFSTA);

	if (cmd->data) {
		int res;
		res = s3c2440sdi_setup_data(host, cmd->data);

		if(res) {
			cmd->error = MMC_ERR_DMA;
			cmd->data->error = MMC_ERR_DMA;

			mmc_request_done(mmc, mrq);
			return;
		}


		if(host->dodma) {
			res = s3c2440sdi_prepare_dma(host, cmd->data);
		} else {
			res = s3c2440sdi_prepare_pio(host, cmd->data);
		}

		if(res) {
			cmd->error = MMC_ERR_DMA;
			cmd->data->error = MMC_ERR_DMA;

			mmc_request_done(mmc, mrq);
			return;
		}

	}

	// Send command
	s3c2440sdi_send_command(host, cmd);

	// Enable Interrupt
	enable_irq(host->irq);
}

/**
 * Write the settings in host->ios to the SDI hardware.
 * Make sure you have obtained a WRITE lock on host->ios_lock prior to this call!
 */
static void s3c2410sdi_write_ios_settings( struct s3c2410sdi_host *host){
	struct mmc_ios *ios;
	u32 sdi_psc, sdi_con;

	ios = &host->ios;

	//Set power
	sdi_con = readl(host->base + S3C2410_SDICON);
	switch(ios->power_mode) {
		case MMC_POWER_ON:
		case MMC_POWER_UP:
#ifdef CONFIG_MACH_TOMTOMGO
			IO_Deactivate(SDCLK);
			IO_SetFunction(SDCMD);
			IO_SetFunction(SDDATA0);
			IO_SetFunction(SDDATA1);
			IO_SetFunction(SDDATA2);
			IO_SetFunction(SDDATA3);

			IO_Activate(SD_PWR_ON);
			IO_SetFunction(SDCLK);
#else
			s3c2410_gpio_cfgpin(S3C2410_GPE5, S3C2410_GPE5_SDCLK);
			s3c2410_gpio_cfgpin(S3C2410_GPE6, S3C2410_GPE6_SDCMD);
			s3c2410_gpio_cfgpin(S3C2410_GPE7, S3C2410_GPE7_SDDAT0);
			s3c2410_gpio_cfgpin(S3C2410_GPE8, S3C2410_GPE8_SDDAT1);
			s3c2410_gpio_cfgpin(S3C2410_GPE9, S3C2410_GPE9_SDDAT2);
			s3c2410_gpio_cfgpin(S3C2410_GPE10, S3C2410_GPE10_SDDAT3);
#endif

			if(!host->is2440) sdi_con|=S3C2410_SDICON_FIFORESET;
			break;

		case MMC_POWER_OFF:
		default:
#ifdef CONFIG_MACH_TOMTOMGO
			IO_Deactivate(SDCLK);
			IO_Deactivate(SDCMD);
			IO_Deactivate(SDDATA0);
			IO_Deactivate(SDDATA1);
			IO_Deactivate(SDDATA2);
			IO_Deactivate(SDDATA3);
			IO_Deactivate(SD_PWR_ON);
#else
			s3c2410_gpio_setpin(S3C2410_GPE5, 0);
			s3c2410_gpio_cfgpin(S3C2410_GPE5, S3C2410_GPE5_OUTP);
#endif
			if(host->is2440) sdi_con|=S3C2440_SDICON_SDRESET;
			break;
	}

	//Set clock
	for(sdi_psc=0;sdi_psc<255;sdi_psc++) {
		host->real_rate = host->clk_rate / (host->clk_div*(sdi_psc+1));
		if(host->real_rate <= ios->clock) break;
	}

	if(sdi_psc > 255) sdi_psc = 255;
	host->prescaler = sdi_psc;

	writel(host->prescaler, host->base + S3C2410_SDIPRE);

	//If requested clock is 0, real_rate will be 0, too
	if(ios->clock == 0) host->real_rate = 0;

	//Set CLOCK_ENABLE
	if(ios->clock) 	sdi_con |= S3C2410_SDICON_CLOCKTYPE;
	else		sdi_con &=~S3C2410_SDICON_CLOCKTYPE;

	writel(sdi_con, host->base + S3C2410_SDICON);

	if((ios->power_mode==MMC_POWER_ON)||(ios->power_mode==MMC_POWER_UP)) {
		CONFDBG("SD/MMC slot is running at %lukHz (requested: %ukHz).\n",
		host->real_rate/1000, ios->clock/1000);
	} else {
		CONFDBG("SD/MMC slot is powered down.\n");
	}
}

static void s3c2410sdi_request(struct mmc_host *mmc, struct mmc_request *mrq) {
 	struct s3c2410sdi_host *host = mmc_priv(mmc);

	host->cmd_is_stop = 0;
	host->mrq = mrq;

	s3c2410sdi_send_request(mmc);
}

static void s3c2410sdi_set_ios(struct mmc_host *mmc, struct mmc_ios *ios) {
	struct s3c2410sdi_host *host = mmc_priv(mmc);

	write_lock_bh( &host->ios_lock );
	{
		/* copy the ios to cache	*/
		memcpy( &host->ios, ios, sizeof( host->ios) );

		/* Write ios in cache to hardware	*/
		s3c2410sdi_write_ios_settings( host );
	}
	write_unlock_bh( &host->ios_lock );
}

static void s3c2410sdi_reset(struct s3c2410sdi_host *host) {
	u32 con = readl(host->base + S3C2410_SDICON);

	con |= S3C2440_SDICON_SDRESET;

	writel(con, host->base + S3C2410_SDICON);
}

static struct mmc_host_ops s3c2410sdi_ops = {
	.request	= s3c2410sdi_request,
	.set_ios	= s3c2410sdi_set_ios,
};

static int s3c24xxsdi_probe(struct device *dev, int have_2440)
{
	struct platform_device	*pdev = to_platform_device(dev);
	struct mmc_host 	*mmc;
	struct s3c2410sdi_host 	*host;

	int ret;

	mmc = mmc_alloc_host(sizeof(struct s3c2410sdi_host), dev);
	if (!mmc) {
		ret = -ENOMEM;
		goto probe_out;
	}

	host = mmc_priv(mmc);

	rwlock_init(&host->ios_lock);

	spin_lock_init(&host->complete_lock);
	tasklet_init(&host->pio_tasklet, pio_tasklet, (unsigned long) host);
	if (have_2440) {
		host->is2440	= 1;
		host->sdiimsk	= S3C2440_SDIIMSK;
		host->sdidata	= S3C2440_SDIDATA;
		host->clk_div	= 1;
	} else {
		host->is2440	= 0;
		host->sdiimsk	= S3C2410_SDIIMSK;
		host->sdidata	= S3C2410_SDIDATA;
		host->clk_div	= 2;
	}
	host->dodma		= 0;
	host->complete_what 	= COMPLETION_NONE;
	host->pio_active 	= XFER_NONE;

	host->mmc 		= mmc;
	host->dma		= S3C2410SDI_DMA;
	host->irq_cd		= IRQ_EINT2;

	mmc->debounce_delay	= 500;

	host->mem = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!host->mem) {
		ERR("failed to get io memory region resouce.\n");
		ret = -ENOENT;
		goto probe_free_host;
	}

	host->mem = request_mem_region(host->mem->start,
		RESSIZE(host->mem), pdev->name);

	if (!host->mem) {
		ERR("failed to request io memory region.\n");
		ret = -ENOENT;
		goto probe_free_host;
	}

	host->base = ioremap(host->mem->start, RESSIZE(host->mem));
	if (host->base == 0) {
		ERR("failed to ioremap() io memory region.\n");
		ret = -EINVAL;
		goto probe_free_mem_region;
	}

	host->irq = platform_get_irq(pdev, 0);
	if (host->irq == 0) {
		ERR("failed to get interrupt resouce.\n");
		ret = -EINVAL;
		goto probe_iounmap;
	}

	if(request_irq(host->irq, s3c2410sdi_irq, 0, DRIVER_NAME, host)) {
		ERR("failed to request sdi interrupt.\n");
		ret = -ENOENT;
		goto probe_iounmap;
	}

	disable_irq(host->irq);

#ifdef CONFIG_MACH_TOMTOMGO
	IO_SetInterruptOnToggle(CD_SD);
#else
	s3c2410_gpio_cfgpin(S3C2410_GPF2, S3C2410_GPF2_EINT2);
	set_irq_type(host->irq_cd, IRQT_BOTHEDGE);
#endif

	if(request_irq(host->irq_cd, s3c2410sdi_irq_cd, 0, DRIVER_NAME, host)) {
		ERR("failed to request card detect interrupt.\n");
		ret = -ENOENT;
		goto probe_free_irq;
	}

	if(s3c2410_dma_request(S3C2410SDI_DMA, &s3c2410sdi_dma_client, NULL)) {
		ERR("unable to get DMA channel.\n");
		ret = -EBUSY;
		goto probe_free_irq_cd;
	}

	host->clk = clk_get(dev, "sdi");
	if (IS_ERR(host->clk)) {
		ERR("failed to find clock source.\n");
		ret = PTR_ERR(host->clk);
		host->clk = NULL;
		goto probe_free_host;
	}

	if((ret = clk_use(host->clk))) {
		ERR("failed to use clock source.\n");
		goto clk_free;
	}

	if((ret = clk_enable(host->clk))) {
		ERR("failed to enable clock source.\n");
		goto clk_unuse;
	}

	host->clk_rate = clk_get_rate(host->clk);

	mmc->ops 	= &s3c2410sdi_ops;
	mmc->ocr_avail	= MMC_VDD_32_33;
	mmc->flags      = MMC_HOST_WIDEMODE;
	mmc->f_min 	= host->clk_rate / (host->clk_div * 256);
	mmc->f_max 	= host->clk_rate / host->clk_div;

	mmc->max_sectors	= 4095;
	mmc->max_seg_size	= mmc->max_sectors << 9;

	mmc->max_phys_segs	= 128;
	mmc->max_hw_segs	= 128;

	DBG("probe: mode:%s mapped sdi_base=%p irq=%u irq_cd=%u dma=%u.\n",
		(host->is2440?"2440":"2410"),
		host->base, host->irq, host->irq_cd, host->dma);

	if((ret = mmc_add_host(mmc))) {
		ERR("failed to add mmc host.\n");
		goto free_dmabuf;
	}

	dev_set_drvdata(dev, mmc);

	INFO("initialisation done.\n");
	return 0;

 free_dmabuf:
	clk_disable(host->clk);

 clk_unuse:
	clk_unuse(host->clk);

 clk_free:
	clk_put(host->clk);

 probe_free_irq_cd:
 	free_irq(host->irq_cd, host);

 probe_free_irq:
 	free_irq(host->irq, host);

 probe_iounmap:
	iounmap(host->base);

 probe_free_mem_region:
	release_mem_region(host->mem->start, RESSIZE(host->mem));

 probe_free_host:
	mmc_free_host(mmc);
 probe_out:
	return ret;
}

static int s3c2410sdi_remove(struct device *dev)
{
	struct mmc_host 	*mmc  = dev_get_drvdata(dev);
	struct s3c2410sdi_host 	*host = mmc_priv(mmc);

	mmc_remove_host(mmc);
	clk_disable(host->clk);
	clk_unuse(host->clk);
	clk_put(host->clk);
 	free_irq(host->irq_cd, host);
 	free_irq(host->irq, host);
	iounmap(host->base);
	release_mem_region(host->mem->start, RESSIZE(host->mem));
	mmc_free_host(mmc);

	return 0;
}

static int s3c2410sdi_probe(struct device *dev)
{
	return s3c24xxsdi_probe(dev, 0);
}

static int s3c2412sdi_probe(struct device *dev)
{
	return s3c24xxsdi_probe(dev, 1);
}

static int s3c2440sdi_probe(struct device *dev)
{
	return s3c24xxsdi_probe(dev, 1);
}

#ifdef CONFIG_PM

static int s3c24XXsdi_suspend(struct device *dev, pm_message_t state, u32 level)
{
	struct mmc_host *mmc = dev_get_drvdata(dev);
	int ret = 0;

	if (mmc && (level == SUSPEND_POWER_DOWN)) {
		ret = mmc_suspend_host(mmc, state);
	}

	return ret;
}

static int s3c24XXsdi_resume(struct device *dev, u32 level)
{
	struct mmc_host *mmc = dev_get_drvdata(dev);
	int ret = 0;

	if (mmc && (level == RESUME_POWER_ON)) {
		ret = mmc_resume_host(mmc);
	}

	return ret;
}

#else /* CONFIG_PM */
#define s3c24XXsdi_suspend NULL
#define s3c24XXsdi_resume NULL
#endif /* CONFIG_PM */


static struct device_driver s3c2410sdi_driver =
{
	.name		= "s3c2410-sdi",
	.bus		= &platform_bus_type,
	.probe		= s3c2410sdi_probe,
	.remove		= s3c2410sdi_remove,
	.suspend	= s3c24XXsdi_suspend,
	.resume		= s3c24XXsdi_resume,
};

static struct device_driver s3c2412sdi_driver =
{
	.name		= "s3c2412-sdi",
	.bus		= &platform_bus_type,
	.probe		= s3c2412sdi_probe,
	.remove		= s3c2410sdi_remove,
	.suspend	= s3c24XXsdi_suspend,
	.resume		= s3c24XXsdi_resume,
};

static struct device_driver s3c2440sdi_driver =
{
	.name		= "s3c2440-sdi",
	.bus		= &platform_bus_type,
	.probe		= s3c2440sdi_probe,
	.remove		= s3c2410sdi_remove,
	.suspend	= s3c24XXsdi_suspend,
	.resume		= s3c24XXsdi_resume,
};


static int __init s3c2410sdi_init(void)
{
	driver_register(&s3c2410sdi_driver);
	driver_register(&s3c2412sdi_driver);
	driver_register(&s3c2440sdi_driver);
	return 0;
}

static void __exit s3c2410sdi_exit(void)
{
	driver_unregister(&s3c2410sdi_driver);
	driver_unregister(&s3c2412sdi_driver);
	driver_unregister(&s3c2440sdi_driver);
}

module_init(s3c2410sdi_init);
module_exit(s3c2410sdi_exit);

MODULE_DESCRIPTION("Samsung S3C24xx Multimedia Card Interface driver");
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Thomas Kleffel <tk@maintech.de>");

