/* drivers/video/tomtom/ttgfb.c
 *
 * TomTom GO LCD Controller Frame Buffer Driver
 * based on skeletonfb.c, sa1100fb.c, tomtomgofb.c
 *
 * Copyright (C) 2005 TomTom BV <http://www.tomtom.com/>
 * Author: Christian Daniel <cd@cdaniel.de>
 * Author: Thomas Kleffel <tk@maintech.de>
 * Author: Jeroen Taverne <jeroen.taverne@tomtom.com>
 * Author: Dimitry Andric <dimitry.andric@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#ifdef __KERNEL__
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/fb.h>
#include <linux/tomtomgofb.h>
#include <linux/dma-mapping.h>

#include <asm/hardware/clock.h>

#include <barcelona/gopins.h>
#include <barcelona/gotype.h>
#include <barcelona/debug.h>
#include <barcelona/Barc_pwm.h>

#define PFX "ttgfb: "
#define PK_DBG PK_DBG_FUNC
#endif /* __KERNEL__ */

#ifdef __BOOTLOADER__
#include "timer.h"
#include "gopins.h"
#include "gotype.h"
#include "option.h"
#define PK_DBG if (0) printf
#define PK_WARN if (0) printf
#define inline __inline
#define mdelay TIMER_Delay
#endif /* __BOOTLOADER__ */

#include "ttgfb.h"
#include "lcdregs.h"
#include "nec.h"
#include "ltv350qv.h"
#include "ltp400wq.h"

static void (*tftInit)(void);
static void (*tftOff)(void);
static unsigned TTGL_XRES;
static unsigned TTGL_YRES;
static unsigned TTGL_XRESV;
static unsigned TTGL_YRESV;
#define TTGL_BPP    16
#define TTGL_BYPP   2
#define TTGL_LINE   (TTGL_XRES * TTGL_BYPP)
#define TTGL_SIZE   (TTGL_YRES * TTGL_XRES * TTGL_BYPP)
#define TTGL_WIDTH  53
#define TTGL_HEIGHT 71
#define TTGL_R_OFS  11
#define TTGL_R_LEN  5
#define TTGL_G_OFS  5
#define TTGL_G_LEN  6
#define TTGL_B_OFS  0
#define TTGL_B_LEN  5
#define TTGL_T_OFS  0
#define TTGL_T_LEN  0
#define TTGL_PAGES  2

#ifdef __KERNEL__
static struct ttgfb_slot fbs;
static struct resource *lcd_mem;
static struct clk *lcd_clock;
#endif

static inline void ttgfb_set_base(unsigned start, unsigned stop)
{
	PK_DBG("Setting viewport from 0x%08x to 0x%08x\n", start, stop);
	rLCDSADDR1 = S3C2410_LCDBANK(start >> 22)|S3C2410_LCDBASEU(start >> 1);
	rLCDSADDR2 = stop >> 1;
}

#ifdef __KERNEL__

/*
 *	s3c2410fb_check_var():
 *	Get the video params out of 'var'. If a value doesn't fit, round it up,
 *	if it's too big, return -EINVAL.
 *
 *	Round up in the following order: bits_per_pixel, xres,
 *	yres, xres_virtual, yres_virtual, xoffset, yoffset, grayscale,
 *	bitfields, horizontal timing, vertical timing.
 */
static int ttgfb_check_var(struct fb_var_screeninfo *var, struct fb_info *info)
{
	PK_DBG("check_var(var=%p, info=%p)\n", var, info);

	if (var->xres > TTGL_XRES)				return -EINVAL;
	if (var->yres > TTGL_YRES)				return -EINVAL;

	if (var->xres_virtual > TTGL_XRESV)		return -EINVAL;
	if (var->yres_virtual > TTGL_YRESV)		return -EINVAL;

	if (var->xoffset != 0)					return -EINVAL;
	if (var->yoffset > TTGL_YRES)			return -EINVAL;

	if (var->bits_per_pixel != TTGL_BPP)	return -EINVAL;
	if (var->grayscale != 0)				return -EINVAL;

	if (var->red.offset   != TTGL_R_OFS)	return -EINVAL;
	if (var->green.offset != TTGL_G_OFS)	return -EINVAL;
	if (var->blue.offset  != TTGL_B_OFS)	return -EINVAL;

	if (var->red.length   != TTGL_R_LEN)	return -EINVAL;
	if (var->green.length != TTGL_G_LEN)	return -EINVAL;
	if (var->blue.length  != TTGL_B_LEN)	return -EINVAL;

	return 0;
}

/**
 *      ttgfb_set_par - Optional function. Alters the hardware state.
 *      @info: frame buffer structure that represents a single frame buffer
 *
 */
static int ttgfb_set_par(struct fb_info *info)
{
	PK_DBG("set_par(info=%p)\n", info);
	return 0;
}


static int ttgfb_setcolreg(unsigned regno, unsigned red, unsigned green, unsigned blue, unsigned transp, struct fb_info *info)
{
	struct ttgfb_slot *slot = (struct ttgfb_slot *) info;
	unsigned bpp, m;

	bpp = info->var.bits_per_pixel;
	m = 1 << bpp;
	if (regno >= m) {
		return -EINVAL;
	}

	switch (bpp) {
	case 16:
		/* RGB 565 */
		slot->pseudo_pal[regno] = ((red & 0xF800)
			| ((green & 0xFC00) >> 5)
			| ((blue & 0xF800) >> 11));
		break;
	}

	return 0;
}

static inline void ttgfb_set_base_wait(unsigned start, unsigned stop)
{
	unsigned long flags;

	local_irq_save(flags);
	/* Wait for retrace, according to vague Samsung sample */
	while ((rLCDCON1 >> 18) == 0)
		;
	ttgfb_set_base(start, stop);
	local_irq_restore(flags);
}

/**
 *	ttgfb_pan_display - NOT a required function. Pans the display.
 *	@var: frame buffer variable screen structure
 *	@info: frame buffer structure that represents a single frame buffer
 *
 *	Pan (or wrap, depending on the `vmode' field) the display using the
 *	`xoffset' and `yoffset' fields of the `var' structure.
 *	If the values don't fit, return -EINVAL.
 *
 *	Returns negative errno on error, or zero on success.
 */
static int ttgfb_pan_display(struct fb_var_screeninfo *var, struct fb_info *info)
{
	unsigned start;
	unsigned stop;

	PK_DBG("xoffset=%u, yoffset=%u\n", var->xoffset, var->yoffset);
	if (var->xoffset != 0)
		return -EINVAL;

	start = info->fix.smem_start + var->yoffset * TTGL_LINE;
	stop = start + TTGL_SIZE;
	ttgfb_set_base_wait(start, stop);

	return 0;
}

/**
 *      xxxfb_blank - NOT a required function. Blanks the display.
 *	@blank_mode: the blank mode we want.
 *	@info: frame buffer structure that represents a single frame buffer
 *
 *	Blank the screen if blank_mode != 0, else unblank. Return 0 if
 *	blanking succeeded, != 0 if un-/blanking failed due to e.g. a
 *	video mode which doesn't support it. Implements VESA suspend
 *	and powerdown modes on hardware that supports disabling hsync/vsync:
 *	blank_mode == 2: suspend vsync
 *	blank_mode == 3: suspend hsync
 *	blank_mode == 4: powerdown
 *
 *	Returns negative errno on error, or zero on success.
 *
 */
static int ttgfb_blank(int blank_mode, struct fb_info *info)
{
	PK_DBG("blank(mode=%d, info=%p)\n", blank_mode, info);
	return 0;
}

static struct fb_ops ttgfb_ops = {
	.owner			= THIS_MODULE,
	.fb_check_var	= ttgfb_check_var,
	.fb_set_par		= ttgfb_set_par,
	.fb_blank		= ttgfb_blank,
	.fb_pan_display	= ttgfb_pan_display,
	.fb_setcolreg	= ttgfb_setcolreg,
	.fb_fillrect	= cfb_fillrect,
	.fb_copyarea	= cfb_copyarea,
	.fb_imageblit	= cfb_imageblit,
	.fb_cursor		= soft_cursor,
};

/* Fake monspecs to fill in fbinfo structure */
/* Don't know if the values are important    */
static struct fb_monspecs ttgfb_monspecs __initdata = {
	.hfmin	= 30000,
	.hfmax	= 70000,
	.vfmin	= 50,
	.vfmax	= 65,
};
#endif /* KERNEL */

#ifdef __KERNEL__
static void ttgfb_setup(struct ttgfb_slot *fbs)
#endif /* KERNEL */
#ifdef __BOOTLOADER__
static void ttgfb_setup(void)
#endif /* __BOOTLOADER__ */
{
#ifdef __KERNEL__
	struct fb_info *fbi= &fbs->fbi;
#endif /* __KERNEL__ */

	switch (IO_GetTftType()) {
	case GOTFT_NEC_NL2432HC22:
		TTGL_XRES = 240;
		TTGL_YRES = 320;
		tftInit = NEC_Init;
		tftOff = NEC_Off;
		break;
	case GOTFT_SAMSUNG_LTV350:
		TTGL_XRES = 320;
		TTGL_YRES = 240;
		tftInit = LTV350QV_Init;
		tftOff = LTV350QV_Off;
		break;
	case GOTFT_SAMSUNG_LTP400:
	case GOTFT_SMDK2440:
		TTGL_XRES = 480;
		TTGL_YRES = 272;
		tftInit = LTP400WQ_Init;
		tftOff = LTP400WQ_Off;
		break;
	default:
		PK_WARN("Unknown TFT type %u, assuming 240x320 resolution.\n", IO_GetTftType());
		TTGL_XRES = 240;
		TTGL_YRES = 320;
		tftInit = 0;
		tftOff = 0;
		break;
	}
	TTGL_XRESV = TTGL_XRES;
	TTGL_YRESV = TTGL_YRES * TTGL_PAGES; /* For page flipping */

#ifdef __KERNEL__
	memset(&fbi->var, 0, sizeof fbi->var);
	fbi->var.xres           = TTGL_XRES;
	fbi->var.yres           = TTGL_YRES;
	fbi->var.xres_virtual   = TTGL_XRESV;
	fbi->var.yres_virtual   = TTGL_YRESV;
	fbi->var.bits_per_pixel = TTGL_BPP;
	fbi->var.red.offset     = TTGL_R_OFS;
	fbi->var.red.length     = TTGL_R_LEN;
	fbi->var.green.offset   = TTGL_G_OFS;
	fbi->var.green.length   = TTGL_G_LEN;
	fbi->var.blue.offset    = TTGL_B_OFS;
	fbi->var.blue.length    = TTGL_B_LEN;
	fbi->var.transp.offset  = TTGL_T_OFS;
	fbi->var.transp.length  = TTGL_T_LEN;
	fbi->var.activate       = FB_ACTIVATE_NOW;
	fbi->var.height         = TTGL_HEIGHT;
	fbi->var.width          = TTGL_WIDTH;
	fbi->var.vmode          = FB_VMODE_NONINTERLACED;

	memset(&fbi->fix, 0, sizeof fbi->fix);
	strcpy(fbi->fix.id, "s3c2410-lcd");
	fbi->fix.smem_len       = TTGL_YRESV * TTGL_XRESV * TTGL_BYPP;
	fbi->fix.type           = FB_TYPE_PACKED_PIXELS;
	fbi->fix.visual         = FB_VISUAL_TRUECOLOR;
	fbi->fix.ypanstep       = 1;
	fbi->fix.line_length    = TTGL_LINE;
	fbi->fix.accel          = FB_ACCEL_NONE;

	fbi->flags              = FBINFO_FLAG_DEFAULT;
	fbi->monspecs           = ttgfb_monspecs;
	fbi->fbops              = &ttgfb_ops;
	fbi->pseudo_palette     = fbs->pseudo_pal;
#endif /* __BOOTLOADER__ */
}

#ifdef __KERNEL__
static int ttgfb_alloc(struct ttgfb_slot *fbs)
{
	fbs->map_size = PAGE_ALIGN(fbs->fbi.fix.smem_len);
	fbs->map_cpu = dma_alloc_writecombine(fbs->dev, fbs->map_size, &fbs->map_dma, GFP_KERNEL);

	if (!fbs->map_cpu)
		return -ENOMEM;

	fbs->fbi.fix.smem_start = fbs->map_dma;
	fbs->fbi.screen_base    = fbs->map_cpu;
	fbs->fbi.screen_size    = fbs->map_size;

	if (!screen_info.lfb_base) {
		/* If no LFB from bootloader, just clear it out. */
		PK_INFO("No LFB from bootloader, clearing screen\n");
		memset(fbs->map_cpu, 0, fbs->map_size);
	} else {
		/* If bootloader supplies LFB, try to copy it. */
		void __iomem *orig_lfb = ioremap(screen_info.lfb_base, fbs->map_size);
		if (!orig_lfb) {
			PK_INFO("Unable to map bootloader LFB, clearing screen\n");
			memset(fbs->map_cpu, 0, fbs->map_size);
		} else {
			u_char *temp;
			PK_INFO("Copying bootloader LFB to all pages\n");
			for (temp = fbs->map_cpu; temp < fbs->map_cpu + fbs->map_size; temp += fbs->map_size / TTGL_PAGES)
				memcpy(temp, orig_lfb, fbs->map_size / TTGL_PAGES);
			iounmap(orig_lfb);
		}
	}
	
	return 0;
}

static void ttgfb_free(struct ttgfb_slot *fbs)
{
	if (fbs->map_cpu) {
		dma_free_writecombine(fbs->dev, fbs->map_size, fbs->map_cpu, fbs->map_dma);
		fbs->map_dma = 0;
		fbs->map_cpu = NULL;
	}
}
#endif /* __KERNEL__ */

#ifdef __BOOTLOADER__
void TFT_Off(void)
#endif
#ifdef __KERNEL__
void tomtomgo_lcdoff(void) // HACK (must be static)
#endif
{
	// Only turn off LCD when it's on
	if (IO_GetInput(LCD_VCC_PWREN))
	{
		// Disable backlight here now to avoid white FLASH
		IO_Deactivate(BACKLIGHT_PWM);

		// Show white screen
		rTPAL = (1<<24)|0xffffff;
		mdelay(100);

		if (tftOff) tftOff();
			
		// Disable timing signals
		IO_Deactivate(LCD_OEN);
		IO_Deactivate(VDEN);
		IO_Deactivate(VCLK);
		IO_Deactivate(HSYNC);
		IO_Deactivate(VSYNC);

		// Stop driving LCD data
		IO_Deactivate(VD0);
		IO_Deactivate(VD1);
		IO_Deactivate(VD2);
		IO_Deactivate(VD3);
		IO_Deactivate(VD4);
		IO_Deactivate(VD5);
		IO_Deactivate(VD6);
		IO_Deactivate(VD7);
		IO_Deactivate(VD8);
		IO_Deactivate(VD9);
		IO_Deactivate(VD10);
		IO_Deactivate(VD11);
		IO_Deactivate(VD12);
		IO_Deactivate(VD13);
		IO_Deactivate(VD14);
		IO_Deactivate(VD15);
		IO_Deactivate(VD16);
		IO_Deactivate(VD17);
		IO_Deactivate(VD18);
		IO_Deactivate(VD19);
		IO_Deactivate(VD20);
		IO_Deactivate(VD21);
		IO_Deactivate(VD22);
		IO_Deactivate(VD23);

		// Turn charge pump off
		IO_Deactivate(LCD_BIAS_PWREN);

		// Wait 50 ms
		mdelay(50);

		// Disable LCD power
		IO_Deactivate(LCD_VCC_PWREN);

		// Wait 50 ms
		mdelay(1200);

#ifdef __KERNEL__
		clk_disable(lcd_clock);
#endif /* __KERNEL__ */
	}
}

#ifdef __KERNEL__
void tomtomgo_lcdon(void) // HACK (must be static 
#endif
#ifdef __BOOTLOADER__
void TFT_On(int memSize)
#endif
{
#ifdef __KERNEL__
	unsigned start = fbs.fbi.fix.smem_start + fbs.fbi.var.yoffset * TTGL_LINE;
	unsigned stop = start + TTGL_SIZE;
	clk_enable(lcd_clock);
#endif /* __KERNEL__ */
#ifdef __BOOTLOADER__
	unsigned start = FRAMEBUFFER_ADDRESS;
	unsigned stop = FRAMEBUFFER_ADDRESS + memSize;
	ttgfb_setup();
//	TFT_Off();
#endif /* __BOOTLOADER__ */

	// Set base frameaddress
	ttgfb_set_base(start, stop);
	// LCDSADDR3 only needs to be initialized once
	rLCDSADDR3 = S3C2410_OFFSIZE(0)|S3C2410_PAGEWIDTH(TTGL_XRES);

	if (!IO_GetInput(LCD_VCC_PWREN))
	{
		// Show black screen
		rTPAL = (1<<24)|0x000000;
	
#if 1
		// Do softstart on certain LCD's
		if (IO_HaveTftSoftStart()) IO_GeneratePWM(LCD_VCC_PWREN);
#else
		IO_Deactivate(LCD_CS);
		
		mdelay(50);
#endif
		// Enable LCD power
		IO_Activate(LCD_VCC_PWREN);
	
		// Wait 50 ms till power is stable
		mdelay(50);
	
		// Turn charge pump on
		IO_Activate(LCD_BIAS_PWREN);
	
		// Wait 50 ms till bias power is stable
		mdelay(50);
	
		// Enable timing signals
		IO_Activate(LCD_OEN);
		IO_SetFunction(VDEN);
		IO_SetFunction(VCLK);
		IO_SetFunction(HSYNC);
		IO_SetFunction(VSYNC);
	
		// Start driving LCD data
		IO_SetFunction(VD0);
		IO_SetFunction(VD1);
		IO_SetFunction(VD2);
		IO_SetFunction(VD3);
		IO_SetFunction(VD4);
		IO_SetFunction(VD5);
		IO_SetFunction(VD6);
		IO_SetFunction(VD7);
		IO_SetFunction(VD8);
		IO_SetFunction(VD9);
		IO_SetFunction(VD10);
		IO_SetFunction(VD11);
		IO_SetFunction(VD12);
		IO_SetFunction(VD13);
		IO_SetFunction(VD14);
		IO_SetFunction(VD15);
		IO_SetFunction(VD18);
		IO_SetFunction(VD19);
		IO_SetFunction(VD20);
		IO_SetFunction(VD21);
		IO_SetFunction(VD22);
		IO_SetFunction(VD23);

		// Do panel dep. controller init
		if (tftInit) tftInit();

		mdelay(100);
		// Show normal screen
		rTPAL = (0<<24);
	}
#ifdef __KERNEL__
	pwm_restore_level();
#endif
}

#ifdef __KERNEL__
static int ttgfb_probe(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct resource *res;
	int size;

	PK_DBG("probe=%p, device=%p\n", pdev, dev);

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (res == NULL) {
		PK_ERR("failed to get memory region resouce\n");
		return -ENOENT;
	}

	size = (res->end-res->start)+1;
	lcd_mem = request_mem_region(res->start, size, pdev->name);
	if (lcd_mem == NULL) {
		PK_ERR("failed to get memory region\n");
		return -ENOENT;
	}

	lcd_clock = clk_get(dev, "lcd");
	if (lcd_clock == NULL) {
		PK_ERR("failed to find lcd clock source\n");
		return -ENOENT;
	}

	clk_use(lcd_clock);

	ttgfb_setup(&fbs);
	ttgfb_alloc(&fbs);
	register_framebuffer(&fbs.fbi);

	/* Init the display */
	tomtomgo_lcdon();

	return 0;
}

static int ttgfb_remove(struct device *dev)
{
	PK_DBG("device=%p\n", dev);

	tomtomgo_lcdoff();
	unregister_framebuffer(&fbs.fbi);
	ttgfb_free(&fbs);
	if (lcd_clock != NULL) {
		clk_disable(lcd_clock);
		clk_unuse(lcd_clock);
		clk_put(lcd_clock);
		lcd_clock = NULL;
	}
	if (lcd_mem != NULL) {
		release_resource(lcd_mem);
		kfree(lcd_mem);
		lcd_mem = NULL;
	}

	return 0;
}

static void ttgfb_shutdown(struct device *dev)
{
	tomtomgo_lcdoff();
}

#ifdef CONFIG_PM

static int ttgfb_suspend(struct device *dev, u32 state, u32 level)
{
	PK_DBG("%s: dev = %p, state = %u, level = %u\n", __func__, dev, state, level);
	if (level == SUSPEND_POWER_DOWN) {
		tomtomgo_lcdoff();
	}
	return 0;
}

static int ttgfb_resume(struct device *dev, u32 level)
{
	PK_DBG("%s: dev = %p, level = %u\n", __func__, dev, level);
	if (level == RESUME_POWER_ON) {
//		tomtomgo_lcdon();	// HACK
	}
	return 0;
}

#else /* CONFIG_PM */
#define ttgfb_suspend NULL
#define ttgfb_resume  NULL
#endif /* CONFIG_PM */

static struct device_driver ttgfb_driver = {
	.name		= "s3c2410-lcd",
	.bus		= &platform_bus_type,
	.probe		= ttgfb_probe,
	.remove		= ttgfb_remove,
	.shutdown	= ttgfb_shutdown,
	.suspend	= ttgfb_suspend,
	.resume		= ttgfb_resume,
};

static int __init ttgfb_init(void) {
	printk(KERN_INFO "TomTom GO Framebuffer Driver, (C) 2005 TomTom BV\n");
	return driver_register(&ttgfb_driver);
}

static void __exit ttgfb_exit(void) {
	driver_unregister(&ttgfb_driver);
}

module_init(ttgfb_init);
module_exit(ttgfb_exit);

MODULE_AUTHOR("Christian Daniel <cd@cdaniel.de>, Thomas Kleffel <tk@maintech.de>, Jeroen Taverne <jeroen.taverne@tomtom.com>, Dimitry Andric <dimitry.andric@tomtom.com> ((c) 2005 TomTom BV)");
MODULE_DESCRIPTION("TomTom GO Framebuffer Driver");
MODULE_LICENSE("GPL");
#endif /* __KERNEL__ */

/* EOF */
